package it.amattioli.applicate.browsing;

import it.amattioli.applicate.context.hibernate.SameContext;
import it.amattioli.applicate.selection.SelectionEvent;
import it.amattioli.applicate.selection.SelectionListener;
import it.amattioli.applicate.selection.SelectionSupport;
import it.amattioli.applicate.selection.Selector;
import it.amattioli.dominate.Entity;
import it.amattioli.dominate.RepositoryRegistry;
import it.amattioli.dominate.Specification;
import it.amattioli.dominate.hibernate.filters.ComparisonType;
import it.amattioli.dominate.specifications.ConjunctionSpecification;
import it.amattioli.dominate.specifications.DisjunctionSpecification;
import it.amattioli.dominate.specifications.StringSpecification;
import it.amattioli.dominate.util.PropertyChangeEmitter;
import it.amattioli.dominate.validation.DefaultValidator;
import it.amattioli.dominate.validation.ValidationResult;
import it.amattioli.dominate.validation.Validator;
import it.amattioli.dominate.validation.ValidationResult.ResultType;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.util.Arrays;
import java.util.List;
import java.util.ResourceBundle;

import org.apache.commons.beanutils.PropertyUtils;
import org.apache.commons.lang.ObjectUtils;
import org.apache.commons.lang.StringUtils;

public class EntitySelector<I extends Serializable, T extends Entity<I>> implements Selector<I,T>, PropertyChangeEmitter {
	private static final String MATCH_PROPERTY = "match";
	private static final String SELECTED_OBJECT_PROPERTY = "selectedObject";
	private static final String DESCRIPTION_PROPERTY = "searchValue";

	public static enum Match {NONE, EXACT, MULTIPLE};
	
	private Class<T> entityClass;
	private String[] searchPropertyNames;
	private T selectedObject;
	private String description;
	private ListBrowser<I,T> entityBrowser;
	private PropertyChangeSupport pChange = new PropertyChangeSupport(this);
	private SelectionListener browserSelectionListener;
	private Match match;
	private PropertyChangeListener selectorListener;
	private Object boundBean;
	private String boundPropertyName;
	private SelectionSupport selectionSupport = new SelectionSupport();
	private Specification<T> additionalSpecification;
	private PropertyChangeListener boundBeanListener;
	private ComparisonType comparisonType = ComparisonType.STARTS;

	public EntitySelector() {
		
	}
	
	public EntitySelector(Class<T> entityClass, final ListBrowser<I,T> placeBrowser, String... searchPropertyNames) {
		this.entityClass = entityClass;
		setEntityBrowser(placeBrowser);
		if (searchPropertyNames.length == 0) {
			this.searchPropertyNames = new String[] {"description"};
		} else {
			this.searchPropertyNames = searchPropertyNames;
		}
	}
	
	public EntitySelector(Class<T> entityClass, String... searchPropertyNames) {
		this(entityClass, new ListBrowserImpl<I, T>(entityClass), searchPropertyNames);
	}
	
	public EntitySelector(Class<T> entityClass, String searchPropertyName, final ListBrowser<I,T> placeBrowser) {
		this(entityClass, placeBrowser, searchPropertyName);
	}
	
	public List<String> getSearchPropertyNames() {
		return Arrays.asList(searchPropertyNames);
	}
	
	public ComparisonType getComparisonType() {
		return comparisonType;
	}

	public void setComparisonType(ComparisonType comparisonType) {
		this.comparisonType = comparisonType;
	}
	
    public void addPropertyChangeListener(final PropertyChangeListener listener) {
        pChange.addPropertyChangeListener(listener);
    }

    public void removePropertyChangeListener(final PropertyChangeListener listener) {
        pChange.removePropertyChangeListener(listener);
    }

    public void firePropertyChange(final String propertyName, final Object oldValue, final Object newValue) {
        pChange.firePropertyChange(propertyName, oldValue, newValue);
    }
	
	public Specification<T> getAdditionalSpecification() {
		return additionalSpecification;
	}

	public void setAdditionalSpecification(Specification<T> additionalSpecification) {
		this.additionalSpecification = additionalSpecification;
		if (entityBrowser != null && additionalSpecification != null) {
			entityBrowser.setSpecification(additionalSpecification);
		}
	}

	public ListBrowser<I,T> getEntityBrowser() {
		return entityBrowser;
	}

	@SameContext
	public void setEntityBrowser(final ListBrowser<I,T> entityBrowser) {
		//this.placeBrowser = LongRunningContext.withCurrentSessionManager(placeBrowser);
		this.entityBrowser = entityBrowser;
		//StringSpecification<T> spec = StringSpecification.newInstance(searchPropertyName);
		//entityBrowser.setSpecification(spec);
		browserSelectionListener = new SelectionListener() {

			public void objectSelected(SelectionEvent event) {
				String old = description;
				T selectedObject = entityBrowser.getSelectedObject();
				if (selectedObject != null) {
					description = getSearchPropertyValue(selectedObject);
					setMatch(Match.EXACT);
				} else {
					description = "";
				}
				firePropertyChange(DESCRIPTION_PROPERTY, old, description);
				setSelectedObject(selectedObject);
			}
			
		};
		entityBrowser.addSelectionListener(browserSelectionListener);
		if (additionalSpecification != null) {
			entityBrowser.setSpecification(additionalSpecification);
		}
	}
	
	private String getSearchPropertyValue(T object) {
		try {
			return (String)PropertyUtils.getProperty(object, searchPropertyNames[0]);
		} catch (IllegalAccessException e) {
			throw new RuntimeException(e);
		} catch (InvocationTargetException e) {
			throw new RuntimeException(e.getCause());
		} catch (NoSuchMethodException e) {
			throw new RuntimeException(e);
		}
	}
	
	private void setSelectedObject(T selectedObject) {
		T old = this.selectedObject;
		this.selectedObject = selectedObject;
		firePropertyChange(SELECTED_OBJECT_PROPERTY, old, selectedObject);
		fireSelectionEvent();
	}

	public T getSelectedObject() {
		return selectedObject;
	}
	
	public void setSelectedObjectId(I id) {
		T old = this.selectedObject;
		I oldId = old == null ? null : old.getId();
		if (!ObjectUtils.equals(oldId, id)) {
			if (id == null) {
				this.selectedObject = null;
				this.description = "";
			} else {
				T selected = RepositoryRegistry.instance().getRepository(entityClass).get(id);
				this.selectedObject = selected;
				this.description = getSearchPropertyValue(selected);
			}
			setMatch(Match.EXACT);
			firePropertyChange(SELECTED_OBJECT_PROPERTY, old, selectedObject);
			firePropertyChange(DESCRIPTION_PROPERTY, old, description);
			fireSelectionEvent();
		}
	}
	
	public I getSelectedObjectId() {
		if (selectedObject == null) {
			return null;
		}
		return selectedObject.getId();
	}

	public String getSearchValue() {
		return description;
	}
	
	private Specification<T> createSpecification(String description, ComparisonType comparisonType) {
		DisjunctionSpecification<T> spec = new DisjunctionSpecification<T>();
		for (String currPropertyName: searchPropertyNames) {
			StringSpecification<T> currSpec = StringSpecification.newInstance(currPropertyName);
			currSpec.setComparisonType(comparisonType);
			currSpec.setValue(description);
			spec.addSpecification(currSpec);
		}
		ConjunctionSpecification<T> completeSpec = new ConjunctionSpecification<T>();
		completeSpec.addSpecification(spec);
		if (additionalSpecification != null) {
			completeSpec.addSpecification(additionalSpecification);
		}
		return completeSpec;
	}

	public void setSearchValue(final String newDescription) {
		String description = StringUtils.upperCase(newDescription);
		String oldDescription = getSearchValue();
		if (!StringUtils.equals(oldDescription, description)) {
			Match oldMatch = getMatch();
			if (StringUtils.isBlank(description)) {
				this.description = "";
				setMatch(Match.EXACT);
				firePropertyChange(DESCRIPTION_PROPERTY, oldDescription, description);
				firePropertyChange(MATCH_PROPERTY, oldMatch, getMatch());
				entityBrowser.select((T)null);
			} else {
				entityBrowser.setSpecification(createSpecification(description, ComparisonType.EXACT));
				if (entityBrowser.getList().size() == 0) {
					entityBrowser.setSpecification(createSpecification(description, comparisonType));
				}
				this.description = description;
				switch (entityBrowser.getList().size()) {
					case 0: 
						setMatch(Match.NONE);
					    break;
					
					case 1:
						setMatch(Match.EXACT);
						break;
						
					default:
						setMatch(Match.MULTIPLE);
					        
				}
				firePropertyChange(DESCRIPTION_PROPERTY, oldDescription, description);
				firePropertyChange(MATCH_PROPERTY, oldMatch, getMatch());
				if (entityBrowser.getList().size() == 1) {
					entityBrowser.select(0);
				}
			}
		}
	}
	
	private void setMatch(Match match) {
		this.match = match;
	}
	
	public Match getMatch() {
		return match;
	}
	
	public void bindTo(final Object bean, final String propertyName) {
		boundBean = bean;
		boundPropertyName = propertyName;
		selectorListener = new PropertyChangeListener() {

			@Override
			public void propertyChange(PropertyChangeEvent evt) {
				if (SELECTED_OBJECT_PROPERTY.equals(evt.getPropertyName())) {
					try {
						PropertyUtils.setProperty(bean, propertyName, evt.getNewValue());
					} catch (IllegalAccessException e) {
						throw new RuntimeException(e);
					} catch (InvocationTargetException e) {
						throw new RuntimeException(e.getCause());
					} catch (NoSuchMethodException e) {
						throw new RuntimeException(e);
					}
				}
			}
			
		};
		addPropertyChangeListener(selectorListener);
		if (boundBean instanceof PropertyChangeEmitter) {
			boundBeanListener = new PropertyChangeListener() {
				
				@Override
				public void propertyChange(PropertyChangeEvent evt) {
					if (evt.getPropertyName().equals(boundPropertyName)) {
						T newValue = (T)evt.getNewValue();
						I id = newValue == null ? null : newValue.getId();
						setSelectedObjectId(id);
					}
				}
			};
			((PropertyChangeEmitter)boundBean).addPropertyChangeListener(boundBeanListener);
		}
		T initial;
		try {
			initial = (T)PropertyUtils.getProperty(bean, propertyName);
		} catch (IllegalAccessException e) {
			throw new RuntimeException(e);
		} catch (InvocationTargetException e) {
			throw new RuntimeException(e.getCause());
		} catch (NoSuchMethodException e) {
			throw new RuntimeException(e);
		}
		if (initial != null) {
			setSelectedObjectId(initial.getId());
		} else {
			setSelectedObjectId(null);
		}
	}
	
	public ValidationResult validate() {
		/*
		if (getMatch().equals(Match.NONE)) {
			String msg = ResourceBundle.getBundle("ValidatorMessages").getString("entitySelector.MatchNone");
			return new ValidationResult(ResultType.INVALID, msg);
		}
		*/
		if (boundBean != null) {
			Validator v = new DefaultValidator(boundBean);
			return v.validateProperty(boundPropertyName, getSelectedObject());
		} else {
			return null;
		}
	}

	@Override
	public void addSelectionListener(SelectionListener listener) {
		selectionSupport.addSelectionListener(listener);
	}

	@Override
	public void removeSelectionListener(SelectionListener listener) {
		selectionSupport.addSelectionListener(listener);
	}
	
	protected void fireSelectionEvent() {
		selectionSupport.notifySelectionListeners(new SelectionEvent(this));
	}
}
