package it.amattioli.applicate.browsing;

import it.amattioli.applicate.commands.CommandEvent;
import it.amattioli.applicate.commands.CommandEventSupport;
import it.amattioli.applicate.commands.CommandListener;
import it.amattioli.applicate.commands.CommandResult;
import it.amattioli.applicate.selection.SelectionEvent;
import it.amattioli.applicate.selection.SelectionListener;
import it.amattioli.applicate.selection.SelectionSupport;
import it.amattioli.dominate.Entity;
import it.amattioli.dominate.Filter;
import it.amattioli.dominate.Repository;
import it.amattioli.dominate.RepositoryRegistry;
import it.amattioli.dominate.Specification;
import it.amattioli.dominate.util.PropertyChangeEmitter;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import it.amattioli.common.properties.PropertyChangeSupport;
import java.io.Serializable;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Standard implementation of the {@link ListBrowser} interface.
 *
 * @author andrea
 *
 * @param <I> Identifier class for the browsing entities
 * @param <T> class of the browsing entities
 */
public class ListBrowserImpl<I extends Serializable, T extends Entity<I>> implements ListBrowser<I, T>, PropertyChangeEmitter {
    private static final Logger logger = LoggerFactory.getLogger(ListBrowserImpl.class);

    private Repository<I, T> repository;
    private Filter filter;
    private Specification<T> specification;
    protected List<T> content = null;
    protected Integer selectedIndex;
    private String orderProperty;
    private boolean reverseOrder;
    private PropertyChangeSupport pChange = new PropertyChangeSupport(this);
    private Class<? extends ObjectBrowser<I, T>> objectBrowserClass;
    private ObjectBrowser<I, T> objectBrowser;
    //private Collection<BrowserSelectionListener> selectionListeners;

    private SelectionSupport selectionSupport = new SelectionSupport();
    private CommandEventSupport cmdEventSupport = new CommandEventSupport();
    private ContentChangeSupport contentChangeSupport = new ContentChangeSupport();

	private PropertyChangeListener specificationChangeListener;

    /**
     * Create a new ListBrowser. Before using it a repository must be set using
     * {@link #setRepository(Repository)}
     */
    protected ListBrowserImpl() {

    }

    /**
     * Create a new ListBrowser given the repository that contains the browsing entities.
     *
     * @param repository the repository that contains the browsing entities
     */
    public ListBrowserImpl(Repository<I, T> repository) {
        this.repository = repository;
    }
    
    /**
     * Create a new repository given the class of the browsing entities.
     * 
     * @param entityClass the class of the browsing entities
     */
    public ListBrowserImpl(Class<T> entityClass) {
    	this(RepositoryRegistry.instance().getRepository(entityClass));
    }

    protected Repository<I, T> getRepository() {
        return this.repository;
    }

    protected void setRepository(Repository<I, T> repository) {
        this.repository = repository;
        invalidateContent();
    }

    protected void invalidateContent() {
        if (content != null) {
        	//if (selectedIndex != null) {
        	//	repository.refresh(content.get(selectedIndex.intValue()));
        	//}
        	for (T curr: content) {
        		repository.refresh(curr);
        	}
            content = null;
            selectedIndex = null;
            firePropertyChange("list", null, null);
            fireContentChange();
            notifySelectionListeners();
        }
    }

    protected boolean isValidContent() {
        return content != null;
    }

    protected void validateContent() {
        if (!isValidContent()) {
            refreshContent();
        }
    }

    protected void refreshContent() {
        if (filter != null) {
        	content = new ArrayList<T>(repository.list(filter));
        } else if (specification != null) {
        	content = new ArrayList<T>(repository.list(specification));
        } else {
        	content = new ArrayList<T>(repository.list());
        }
        if (!isValidSelectionIndex(selectedIndex)) {
        	selectedIndex = null;
        }
    }

    @Override
    public void fillCollection(Collection<? super T> toBeFilled) {
		if (isValidContent()) {
			toBeFilled.addAll(content);
		} else {
			content = new ArrayList<T>();
			if (filter == null) {
				repository.fillCollection(content, toBeFilled);
			} else {
				repository.fillCollection(filter, content, toBeFilled);
			}
		}
	}

    @Override
    public List<T> getList() {
        validateContent();
        return content;
    }

    public boolean isEmptyList() {
        return getList().size() == 0;
    }

    @Override
    public void setFilter(final Filter filter) {
        this.filter = filter;
        filter.addPropertyChangeListener(new PropertyChangeListener() {

            public void propertyChange(final PropertyChangeEvent evt) {
                invalidateContent();
            }

        });
        invalidateContent();
    }

    @Override
    public Filter getFilter() {
        return this.filter;
    }

    @Override
	public Specification<T> getSpecification() {
		return this.specification;
	}

	@Override
	public void setSpecification(Specification<T> spec) {
		this.specification = spec;
		specificationChangeListener = new PropertyChangeListener() {

            public void propertyChange(final PropertyChangeEvent evt) {
                invalidateContent();
            }

        };
		spec.addPropertyChangeListener(specificationChangeListener);
		invalidateContent();
	}

	@Override
    public void setOrder(final String property, final boolean reverse) {
        if (!StringUtils.equals(property, this.orderProperty) || reverse != this.reverseOrder) {
            T selected = getSelectedObject();
            this.orderProperty = property;
            this.reverseOrder = reverse;
            if (repository != null) {
                repository.setOrder(property, reverse);
            }
            invalidateContent();
            if (selected != null) {
                select(selected);
            }
        }
    }

    @Override
    public void setOrder(final String property) {
        if (StringUtils.equals(property, getOrderProperty())) {
            setOrder(property, !this.reverseOrder);
        } else {
            setOrder(property, false);
        }
    }

    @Override
    public String getOrderProperty() {
        return this.orderProperty;
    }

    @Override
    public boolean getReverseOrder() {
        return this.reverseOrder;
    }

    @Override
    public void select(int index) {
        select(Integer.valueOf(index));
    }

    public void select(final Integer index) {
    	if (!isValidSelectionIndex(index)) {
    		throw new ArrayIndexOutOfBoundsException(index);
    	}
        selectedIndex = index;
        notifySelectionListeners();
    }

	protected boolean isValidSelectionIndex(final Integer index) {
		return index == null || index < getList().size();
	}

    @Override
    public void select(final T object) {
        if (object == null || isEmptyList()) {
            select((Integer) null);
        } else {
            select(0);
            do {
                if (getSelectedObject().equals(object)) {
                    return;
                }
                next();
            } while (getHasNext());
        }
        select((Integer) null);
    }

    @Override
    public void deselect() {
        select((Integer) null);
    }

    protected T refreshObject(final int index) {
        // Entity toBeRefreshed =
        // (Entity)getList().get(selectedIndex.intValue());
    	T toBeRefreshed = getList().get(index);
        if (toBeRefreshed.getId() != null) {
	        toBeRefreshed = repository.get(toBeRefreshed.getId());
	        toBeRefreshed.getId(); // Costringe l'inizializzazione di un oggetto proxied
	        getList().set(index, toBeRefreshed);
        }
        return toBeRefreshed;
    }

    @Override
    public T getSelectedObject() {
        if (selectedIndex == null) {
            return null;
        }
        // Entity selected = (Entity)getList().get(selectedIndex.intValue());
        // selected = repository.get(selected.getId());
        // getList().set(selectedIndex.intValue(), selected);
        T selected = refreshObject(selectedIndex.intValue());
        return selected;
    }

    @Override
    public Integer getSelectedIndex() {
        return selectedIndex;
    }

    @Override
    public void next() {
        if (getHasNext()) {
            select(getSelectedIndex().intValue() + 1);
        }
    }

    @Override
    public boolean getHasNext() {
        return (getSelectedIndex() != null) && (getSelectedIndex().intValue() < (getList().size() - 1));
    }

    @Override
    public void previous() {
        if (getHasPrevious()) {
            select(getSelectedIndex().intValue() - 1);
        }
    }

    @Override
    public boolean getHasPrevious() {
        return (getSelectedIndex() != null) && (getSelectedIndex().intValue() > 0);
    }

    @Override
    public void commandDone(final CommandEvent source) {
        Integer idx = getSelectedIndex();
        invalidateContent();
        if (isValidSelectionIndex(idx)) {
        	select(idx);
        }
        cmdEventSupport.fireCommandEvent(source);
    }

    protected void addCommandListener(final CommandListener listener) {
        cmdEventSupport.addListener(listener);
    }

    protected void addCommandListener(final CommandListener listener, final CommandResult... results) {
        cmdEventSupport.addListener(listener, results);
    }
    
    protected void fireCommandEvent(final CommandEvent source) {
    	cmdEventSupport.fireCommandEvent(source);
    }

    public void useObjectBrowserClass(Class<? extends ObjectBrowserImpl<I, T>> placeHolderClass) {
        this.objectBrowserClass = placeHolderClass;
    }

    private Constructor<? extends ObjectBrowserImpl<I, T>> findPlaceHolderConstructor() {
        for (Constructor<?> constructor : objectBrowserClass.getConstructors()) {
            Class<?>[] parms = constructor.getParameterTypes();
            if (parms.length == 2 && Entity.class.isAssignableFrom(parms[0])
                    && Repository.class.isAssignableFrom(parms[1])) {
                return (Constructor<? extends ObjectBrowserImpl<I, T>>)constructor;
            }
        }
        return null;
    }

    @Override
    public ObjectBrowser<I, T> newObjectBrowser() {
        ObjectBrowserImpl<I, T> result;
        if (objectBrowserClass != null) {
            Constructor<? extends ObjectBrowserImpl<I, T>> constructor;
            try {
                constructor = findPlaceHolderConstructor();
            } catch (SecurityException e) {
                throw new RuntimeException(e);
            }
            try {
                result = constructor.newInstance(new Object[] {getSelectedObject(), getRepository()});
            } catch (IllegalArgumentException e) {
                throw new RuntimeException(e);
            } catch (InstantiationException e) {
                throw new RuntimeException(e);
            } catch (IllegalAccessException e) {
                throw new RuntimeException(e);
            } catch (InvocationTargetException e) {
                throw new RuntimeException(e.getCause());
            }
        } else {
        	result = new ObjectBrowserImpl<I, T>(getSelectedObject(), getRepository());
        }
        cmdEventSupport.addListener(result);
        return result;
    }

    @Override
    public ObjectBrowser<I, T> getSelectedObjectBrowser() {
        try {
            if (objectBrowser == null) {
                objectBrowser = newObjectBrowser();
                addSelectionListener(objectBrowser);
            }
        } catch (Exception e) {
            logger.error("Errore", e);
            throw new RuntimeException(e);
        }
        return objectBrowser;
    }

    @Override
    public void addPropertyChangeListener(PropertyChangeListener listener) {
        pChange.addPropertyChangeListener(listener);
    }

    @Override
    public void removePropertyChangeListener(PropertyChangeListener listener) {
        pChange.removePropertyChangeListener(listener);
    }
    
    protected void firePropertyChange(String propertyName, Object oldValue, Object newValue) {
		pChange.firePropertyChange(propertyName, oldValue, newValue);
	}

    @Override
    public void addSelectionListener(SelectionListener listener) {
    	selectionSupport.addSelectionListener(listener);
    }

    @Override
    public void removeSelectionListener(SelectionListener listener) {
    	selectionSupport.removeSelectionListener(listener);
    }

    private void notifySelectionListeners() {
    	selectionSupport.notifySelectionListeners(new SelectionEvent(this));
    }

	@Override
	public void release() {
		pChange.disable();
		contentChangeSupport.disable();
	}

	@Override
	public void addContentChangeListener(ContentChangeListener listener) {
		contentChangeSupport.addContentChangeListener(listener);
	}

	@Override
	public void removeContentChangeListener(ContentChangeListener listener) {
		contentChangeSupport.removeContentChangeListener(listener);
	}
	
	protected void fireContentChange() {
		contentChangeSupport.notifyContentChangeListeners(new ContentChangeEvent(this));
	}

	public String toString() {
		return "ListBrowserImpl [repository=" + repository + "]";
	}
}
